@@ -12,3 +12,8 @@ from contextualized.analysis.effects import (
     plot_homogeneous_predictor_effects,
     plot_heterogeneous_predictor_effects,
 )
+from contextualized.analysis.pvals import (
+    calc_homogeneous_context_effects_pvals,
+    calc_homogeneous_predictor_effects_pvals,
+    calc_heterogeneous_predictor_effects_pvals,
+)
